/* Simusparc writen by Kenji Kise 1998,1999,2000,2001
 * This software is provides with absolutely no warrantly.
 */
#include "define.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
#include <signal.h>
#include <fcntl.h>     /* open */
#include <memory.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <utime.h>
#include <limits.h>
#include <unistd.h>
#include <assert.h>

static char UsageMessage[] =
  "SimuSparc version 0.9a last updated Jan 2001\n\
Usage: SimuSparc [-option] command\n\
 -h: help. print this message.\n\
 -d: debug mode\n\
 -e[NUM][m]: simu ends the execution at specified TC.\n\
   m stands for million, ex. 10m means 10 million instructions.\n\
";


/* global variables */
char          *mem_temp;  /* temporary mem area*/
unsigned int  ir;         /* instruction at pc */
unsigned int  r[32];      /* integer  register */
float         f[33];      /* floating register */
unsigned int  y;          /* y register, for MULScc etc. */

unsigned int  n;          /* N-Negative, icc Bit23 of PSR */
unsigned int  z;          /* Z-Zero    , icc Bit22 of PSR */
unsigned int  v;          /* V-Overflow, icc Bit21 of PSR */
unsigned int  c;          /* C-Carry   , icc Bit20 of PSR */

unsigned int  psr;        
unsigned int  fsr;        /* */
unsigned int  pc, npc;    /* program counter, next  */
unsigned int  annul;      /* annul bit by XXicc inst*/

unsigned int  inst_sort;  /* instruntion sorts, instruction.h */
int           toend;      /* exit syscall valid this */
int           reg_size;   /* for save and restore */
unsigned int  argv_argp_count; /* for stack */

unsigned int  option_flag;   /*for option*/
unsigned int  value1,value2; /* value of loads and stores */

unsigned int  count_outfile1; /* for outfile1 format */
ULL     end_tc=0;    /* outfile1 */
ULL     start_tc=0;  /* outfile1 */
unsigned int  memory_area;          /* segmentation fault check */
FILE *outfile;
FILE *new_stdout;
FILE *gfile; /* gziped trace file */
ULL total_c;    /* in main_loop etc. */

extern int fpops();
extern int fcmps();
extern void debugmode();   /* -d option is added */
extern void syscallmode(); /* -s option is added */
extern int Ticc(); 

extern void initialize_memory(int argc, char **argv, char **argp);
extern unsigned int load_word(unsigned int address);
extern void store_word(unsigned int data, unsigned int address);
extern unsigned int mem_address(unsigned int address);
extern void store_byte(unsigned char data, unsigned int address);
  
void usage(){
  printf("%s", UsageMessage);
  exit(0);
}

int destruct_data(int argc, char **argv){
  fprintf(stderr,"** %lld instructions are executed.\n", total_c);
  fprintf(stderr,"** Simulation finished normally.\n");
  return 0;
}  

int construct_data(int argc, char **argv, char **argp){
  int i;
  int value;
  
  time_t time_tmp;
  unsigned long long temp=0;
  char *buf;
  
  buf = (char *)malloc(1024);
  y=n=v=c=fsr=0; /* initialize flag */
  z=1;
  
  option_flag = 0;  /* options are parsed here */
  if(argc>1){
    for(i=0; i<10; i++){  /* at most ten options are available */
      if(argc==1) break;
      if((*(char *)(argv[1]))=='-'){
	switch( *((char *)(argv[1]) +1) )
	  {
	  case 'h': usage();                           break;
	  case 'd': option_flag |= DEBUG_FLAG;         break;
	  case 'i': option_flag |= TICC_FLAG;          break;
	  case 'f': option_flag |= STDOUT_TO_FILE;     break;
	  case 'c': option_flag |= SYSTEMCALL_FLAG;    break;
 	  case 't': option_flag |= DISTRIBUTION_FLAG;  break;
	    
	  case 'e': /* end of simusparc */
	    end_tc =atoll((char *)(argv[1])+2);
	    if(*((char *)(argv[1])+strlen(argv[1])-1)=='m')
	      end_tc *=1000000;
	    fprintf(stderr,"Specified end TC           : %lld\n",
		    end_tc);
	    break;

	  default:
	    fprintf(stderr, "Perhaps illegal options are used.\n");
	    fprintf(stderr, "See the manual.\n");
	    exit(1);
	  }
	argv++;
	argc--;
      }
    }
  }
  if(argc < 2) usage();

  for(i=0;i<32;i++) {r[i]=0; f[i]=0;}
  initialize_memory(argc, argv, argp); /* important */
  if((mem_temp = (char *)malloc(MEM_TEMP_SIZE))==0){  
    fprintf(OUTFILE1,"malloc error for mem_temp[]\n"); exit(1);}
  /* memory temp this is used for system call */
  npc = pc+4;

  return 0;
}


/*
 * From now on, definitions of instructions follow.
 ************************************************************/
int call(){
  unsigned int disp30;
  unsigned int t;
  
  inst_sort=_CALL_;
  
  disp30=(ir & DISP30);
  t = pc;
  r[15]=pc;
  pc = npc;
  npc = t + disp30*4;
  
  value1 = r[15]; /* for outfile */
  return 0;
}

int Bicc(unsigned int ir){
  unsigned int cond;
  unsigned int imm22;
  unsigned int ttrue;
  unsigned int pc_temp;
  int      sign=0;
  int      rss;
  int      a;

  inst_sort=_Bicc_;
  
  a=( ir & A )>>29;
  cond=( ir & COND)>>25;
  imm22= ir & IMM22 ;

  ttrue = 0;
  switch(cond)
    {
    case 0:       ttrue=0; break;           /*bn*/
    case 1:  if(z)ttrue=1; break;           /*be */
    case 2:  if(z | (n^v)) ttrue=1; break;  /*ble*/
    case 3:  if(n^v)ttrue=1;break;          /*bl */
    case 4:  if(c|z)ttrue=1;break;          /*bleu*/
    case 5:  if(c)ttrue=1; break;           /*bcs*/
    case 6:  if(n)ttrue=1; break;           /*bneg*/
    case 7:  if(v)ttrue=1; break;           /*bvs*/
    case 8:       ttrue=1; break;           /*ba*/
    case 9:  if(z==0)ttrue=1; break;        /*bne*/
    case 10: if(!(z | (n^v)))ttrue=1; break;/*bg*/
    case 11: if(!(n^v))ttrue=1; break;      /*bge*/
    case 12: if(!(c|z))ttrue=1; break;      /*bgu*/
    case 13: if(!c)ttrue=1; break;          /*bcc*/
    case 14: if(!n)ttrue=1; break;          /*bpos*/
    case 15: if(!v)ttrue=1; break;          /*bvc*/
    default: fprintf(stderr,"bicc unknown default in branch.cc\n");
      exit(1);
    }

  sign=(imm22 & BIT21) >>21;
  if(sign) rss = (ir & IMM22) | EXTND22;
  else rss = ir & IMM22;
  pc_temp = pc;

  pc=npc;
  if(ttrue){
    npc = pc_temp+rss*4;
    if(cond==8 && a==1)
      annul = 1;
  }
  else{
    npc +=4;
    if(a==1)
      annul = 1;
  }

  return 0;
}

int FBfcc(unsigned int ir){
  unsigned int cond;
  unsigned int imm22;
  unsigned int ttrue;
  unsigned int pc_temp;
  int      sign=0;
  int      rss;
  int      a;
  unsigned int temp,E,L,G,U;

  inst_sort=_FBfcc_;
  
  a=( ir & A )>>29;
  imm22= ir & IMM22 ;
  temp = (fsr>>10) & 0x3;

  E = (temp==0);
  L = (temp==1);
  G = (temp==2);
  U = (temp==3);

  ttrue = 0;
  cond=( ir & COND)>>25;
  switch(cond)
    {
    case 0:                ttrue=0; break; /*fbn*/
    case 1: if(U || L || G)ttrue=1; break; /*fbne*/
    case 2: if(L || G)     ttrue=1; break; /*fblg*/
    case 3: if(U || L)     ttrue=1; break; /*fbul*/
    case 4: if(L)          ttrue=1; break; /*fbl*/
    case 5: if(U || G)     ttrue=1; break; /*fbug*/
    case 6: if(G)          ttrue=1; break; /*fbg*/
    case 7: if(U)          ttrue=1; break; /*fbu*/
    case 8:                ttrue=1; break; /*fba*/
    case 9: if(E)          ttrue=1; break; /*fbe*/
    case 10:if(U || E)     ttrue=1; break; /*fbue*/
    case 11:if(G || E)     ttrue=1; break; /*fbge*/
    case 12:if(U || G || E)ttrue=1; break; /*fbuge*/
    case 13:if(L || E)     ttrue=1; break; /*fble*/
    case 14:if(U || L || E)ttrue=1; break; /*fbule*/
    case 15:if(L || G || E)ttrue=1; break; /*fbo*/
    default:fprintf(OUTFILE1,"fbicc unknown default\n");exit(1);
    }

  sign=(imm22 & BIT21) >>21;
  if(sign) rss = (ir & IMM22) | EXTND22;
  else rss = ir & IMM22;
  pc_temp = pc;

  pc=npc;
  if(ttrue){
    npc = pc_temp+rss*4;
    if(cond==8 && a==1)
      annul = 1;
  }
  else{
    npc +=4;
    if(a==1)
      annul = 1;
  }

  return 0;
}

int SETHI(unsigned int ir){
  unsigned rd;
  unsigned imm22;
  
  inst_sort=_SETHI_;
  
  rd=( ir & RD )>>25;
  imm22=ir & IMM22;

  if(rd!=0){
    r[rd]=imm22<<10;
  }
  value1 = r[rd]; /* for outfile */
  return 0;
}


int CBccc(){
  inst_sort=_CBccc_;
  fprintf(OUTFILE1,"CBfcc is used. \n");
  fprintf(stderr, "CBfcc is used. \n");
  exit(1);
}


int ar_log(){
  unsigned int op3;
  unsigned int result;
  unsigned int opt1;
  unsigned int opt2;
  unsigned int rss;
  int shift;
  int msb;
  unsigned int rd,rs1,rs2;
  unsigned int simm13;
  unsigned int i;
  
  rd  = ( ir & RD )>>25;
  rs1 = ( ir & RS1)>>14;
  rs2 = ir & RS2;
  
  simm13=ir & SIMM13;
  i =( ir & I)>>13;
  
  if(i){         /* get rss by simm13*/
    if( (simm13 & BIT12)>>12 ) rss=(ir & SIMM13) | EXTND13;
    else rss=ir & SIMM13;
  }
  else rss=r[rs2];
  
  op3 = (ir & OP3)>>19;

  switch(op3)
    {
    case 0 : r[rd]=r[rs1] + rss;   inst_sort=_ADD_;  value1=r[rd]; break;
    case 1 : r[rd]=r[rs1] & rss;   inst_sort=_AND_;  value1=r[rd]; break;
    case 2 : r[rd]=r[rs1] | rss;   inst_sort=_OR_;   value1=r[rd]; break;
    case 3 : r[rd]=r[rs1] ^ rss;   inst_sort=_XOR_;  value1=r[rd]; break;
    case 4 : r[rd]=r[rs1] - rss;   inst_sort=_SUB_;  value1=r[rd]; break;
    case 5 : r[rd]=r[rs1] & (~rss);inst_sort=_ANDN_; value1=r[rd]; break;
    case 6 : r[rd]=r[rs1] | (~rss);inst_sort=_ORN_;  value1=r[rd]; break;
    case 7 : r[rd]=r[rs1] ^ (~rss);inst_sort=_XNOR_; value1=r[rd]; break;
    case 8 : r[rd]=r[rs1] + rss +c;inst_sort=_ADDX_; value1=r[rd]; break; 
    case 12: r[rd]=r[rs1] - rss -c;inst_sort=_SUBX_; value1=r[rd]; break;

    case 0x10:      inst_sort=_ADDcc_;
      result=r[rs1] + rss;
      n=bit31(result);
      if(result==0) z=1; else z=0;
      v=(bit31(r[rs1]) & bit31(rss) & !bit31(result) ) |
	(!bit31(r[rs1]) & !bit31(rss) & bit31(result) );
      c=( bit31(r[rs1]) & bit31(rss) ) | 
	( !bit31(result) & ( bit31(r[rs1]) | bit31(rss) ) );
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      break;
      
    case 0x11:      inst_sort=_ANDcc_;
      result=r[rs1] & rss;		    
      n=bit31(result);
      if(result==0) z=1; else z=0;
      v=0;
      c=0;
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      break;
      
    case 18  :      inst_sort=_ORcc_;
      r[rd]=r[rs1] | rss;		    
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=0;
      c=0;
      value1=r[rd]; /* for outfile */
      break;
      
    case 19  :      inst_sort=_XORcc_;
      r[rd]=r[rs1] ^ rss;		    
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=0;
      c=0;
      value1=r[rd]; /* for outfile */
      break;
      
    case 20:        inst_sort=_SUBcc_;
      result=r[rs1]-rss;
      n=bit31(result);
      if(result==0) z=1; else z=0;
      v=(bit31(r[rs1]) & !bit31(rss) & !bit31(result) ) |
	(!bit31(r[rs1]) & bit31(rss) & bit31(result) );
      c=( !bit31(r[rs1]) & bit31(rss) ) | 
	( bit31(result) & (!bit31(r[rs1]) | bit31(rss)));
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      break;
      
    case 21:        inst_sort=_ANDNcc_;
      r[rd]=r[rs1] & (~rss);
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=0;
      c=0;
      value1=r[rd]; /* for outfile */
      break;
      
    case 22  :      inst_sort=_ORNcc_;
      r[rd]=r[rs1] | (~rss) ;		    
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=0;
      c=0;
      value1=r[rd]; /* for outfile */
      break;
      
    case 23  :      inst_sort=_XNORcc_;
      r[rd]=r[rs1] ^ (~rss) ;		    
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=0;
      c=0;
      value1=r[rd]; /* for outfile */
      break;
      
    case 24:        inst_sort=_ADDXcc_;
      result=r[rs1]+rss+c;
      n=bit31(r[rd]);
      if(result==0)z=1; else z=0;
      v=(bit31(r[rs1]) & bit31(rss) & !bit31(result) ) |
	(!bit31(r[rs1]) & !bit31(rss) & bit31(result) );
      c=( bit31(r[rs1]) & bit31(rss) ) | 
	( !bit31(result) & ( bit31(r[rs1]) | bit31(rss) ) );
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      break;
      
    case 28:        inst_sort=_SUBXcc_;
      result=r[rs1]-rss-c;
      n=bit31(r[rd]);
      if(r[rd]==0) z=1; else z=0;
      v=(bit31(r[rs1]) & !bit31(rss) & !bit31(result) ) |
	(!bit31(r[rs1]) & bit31(rss) & bit31(result) );
      c=( !bit31(r[rs1]) & bit31(rss) ) | 
	( bit31(result) & (!bit31(r[rs1]) | bit31(rss)));
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      break;
      
    case 32:        inst_sort=_TADDcc_;
      fprintf(OUTFILE1, "used TADDcc instruction \n ") ;
      exit(1);
      break;
      
    case 33:        inst_sort=_TSUBcc_;
      fprintf(OUTFILE1, "used TSUBcc instruction \n") ;
      exit(1);
      break;
      
    case 34:        inst_sort=_TADDccTV_;
      fprintf(OUTFILE1," used TADDccTV instruvtion \n") ;
      exit(1);
      break;
      
    case 35:        inst_sort=_TSUBccTV_;
      fprintf(OUTFILE1," used TSUBccTV instruction \n") ;
      exit(1);
      break;
      
    case 36:        inst_sort=_MULScc_;
      if(n^v) opt1=BIT31 | (r[rs1]>>1); /*step 3*/
      else    opt1= r[rs1]>>1;
      /*      flag = (y & 1) ;  */ /* for outfile1 */
      if((y & 1)==1)opt2= rss;       /*step 2*/
      else     opt2=0;
      result=opt1+opt2;
      if(r[rs1]&1) y=(y>>1)|BIT31;     /*step 3*/
      else            y=y>>1;
      n=bit31(r[rd]);              
      if(result==0) z=1; else z=0;
      v=((bit31(opt1) & bit31(opt2) & !bit31(result)) |
	 (!bit31(opt1) & !bit31(opt2) & bit31(result))) ;
      c= ( bit31(opt1) & bit31(opt2) ) |
	  (!bit31(result) & (bit31(opt1) | bit31(opt2)) );
      r[rd]=result;
      value1=r[rd]; /* for outfile */
      value2=y;     /* for outfile */
      break;
      
    case 37:        inst_sort=_SLL_;
      if(i)  shift= simm13 & RS2 ;
      else   shift=r[rs2] & RS2 ;
      r[rd]=r[rs1] << shift;
      value1=r[rd]; /* for outfile */
      break;
      
    case 38:        inst_sort=_SRL_;
      if(i)  shift= simm13 & RS2 ;
      else   shift=r[rs2] & RS2 ;
      r[rd]=r[rs1] >> shift;
      value1=r[rd]; /* for outfile */
      break;
      
    case 39:        inst_sort=_SRA_;
      if(i)  shift= simm13 & RS2 ;
      else   shift=r[rs2] & RS2 ;
      result=r[rs1] >> shift;
      msb= r[rs1] >>31;
      r[rd]=result;
      if(msb)r[rd]=r[rd]|( 0xffffffff<<(32-shift) );
      else r[rd]=r[rd];
      value1=r[rd]; /* for outfile */
      break;
      
    case 40:        inst_sort=_RDY_;
      r[rd]=y;
      value1=r[rd]; /* for outfile */
      break;
      
    case 48:        inst_sort=_WRY_;
      y=r[rs1]^rss;
      value1=y; /* for outfile */
      break;
      
    case 0x34:
      fpops(ir);
      break;
      
    case 0x35:
      fcmps(ir);
      break;
      
    case 56:        inst_sort=_JMPL_;
      r[rd] = pc;
      pc = npc;
      npc = r[rs1] + rss;
      
      value1=r[rd]; /* for outfile */
      break;
      
    case 58:        inst_sort=_Ticc_;
      Ticc( (r[rs1]+rss) & USGNB );
      break;
      
    case 60:        inst_sort=_SAVE_;
      for(i=0; i<16; i++) store_word(r[i+16], r[14]+4*i);
      /* store locals and outs registers into memory */
      result=r[rs1]+rss;
      for(i=24; i<32; i++) r[i]=r[i-16]; 
      for(i=8;  i<16; i++) r[i]=0;       /* set 0 outs registers */
      for(i=16; i<24; i++) r[i]=load_word(result+4*i); 
      r[rd]=result;
      break;
      
    case 61:        inst_sort=_RESTORE_;
      for(i=0; i<16; i++) store_word(r[i+16], r[14]+4*i);
      result=r[rs1]+rss;
      for(i=8; i<16; i++) r[i]=r[i+16];
      r[rd]=result;
      for(i=0; i<16; i++) r[i+16]=load_word(r[14]+4*i);
      break;
      
    case 59:      inst_sort=_IFLUSH_;
      fprintf(stderr," used IFLUSH instruction \n") ;
      exit(1);
      break;
      
    default:
      fprintf(stderr,"\nError ar_log.");
      exit(1);
    }
  return 0;
}

unsigned int ld_st(){
  unsigned int rd,rs1,rs2 ;
  unsigned int op3;
  unsigned int i, rss, simm13;
  unsigned int temp_swap, word_swap;
  unsigned char byte;
  unsigned short halfwd;
  unsigned int sign;
  unsigned int mem_adr;
  
  rd=( ir & RD )>>25;
  rs1=( ir & RS1)>>14;
  rs2= ir & RS2;

  simm13=ir & SIMM13;
  i=( ir & I)>>13;
  
  if(i){ /* get rss by simm13 */
    if((simm13 &BIT12)>>12) rss=(ir & SIMM13) | EXTND13;
    else rss=ir & SIMM13;
  }
  else rss=r[rs2];

  mem_adr=r[rs1]+rss;

  op3 = (ir & OP3)>>19;
  switch(op3)
    {
    case 0  :       inst_sort=_LD_;
      memcpy(&r[rd], (char *)mem_address(mem_adr), 4);
      value1 = r[rd]; /* for debug */
      break;
      
    case  1 :       inst_sort=_LDUB_;
      memcpy(&byte, (char *)mem_address(mem_adr), 1);
      r[rd]= (unsigned int)byte;
      value1 = r[rd]; /* for debug */
      break;
      
    case 2 :       inst_sort=_LDUH_;
      memcpy(&halfwd, (char *)mem_address(mem_adr), 2);
      r[rd]= (unsigned int)halfwd ;
      value1 = r[rd]; /* for debug */
      break;
      
    case  3 :      inst_sort=_LDD_;
      memcpy(&r[rd],(char *)mem_address(mem_adr),8);
      value1 = r[rd];   /* for debug */
      value2 = r[rd+1]; /* for debug */
      break;
      
    case  9 :      inst_sort=_LDSB_;
      memcpy(&byte, (char *)mem_address(mem_adr), 1);
      sign=(byte & BIT7)>>7;
      if(sign) r[rd]= (unsigned int)byte | EXTND8;
      else r[rd]= (unsigned int)byte;
      value1 = r[rd];   /* for debug */
      break;
      
    case 10 :      inst_sort=_LDSH_;
      memcpy(&halfwd, (char *)mem_address(mem_adr), 2);
      sign=(halfwd & BIT15)>>15;
      if(sign) r[rd] = (unsigned int)halfwd | EXTND16;
      else r[rd]= (unsigned int)halfwd ;
      value1 = r[rd];   /* for debug */
      break;
      
    case 13 :      inst_sort=_LDSTUB_;
      memcpy(&byte, (char *)mem_address(mem_adr), 1);
      r[rd]= (unsigned int)byte;
      value1 = r[rd];   /* for debug */
      store_byte((unsigned char)0xff, mem_adr);
      break;
      
    case 32 :      inst_sort=_LDF_;
      memcpy(&f[rd],(char *)mem_address(mem_adr),4);
      value1 = f[rd];   /* for debug */
      break;
      
    case 35 :      inst_sort=_LDDF_;
      memcpy(&f[rd],  (char *)mem_address(mem_adr),8);
      value1 = f[rd];   /* for debug */
      value2 = f[rd+1]; /* for debug */
      break;
      
    case 33 :      inst_sort=_LDFSR_;
      memcpy(&fsr, (char *)mem_address(mem_adr), 4);
      value1 = fsr;   /* for debug */
      break;

      
    case  4 :      inst_sort=_ST_; /* Store Word */
      memcpy((char *)mem_address(mem_adr), &r[rd], 4); 
      break;
      
    case  5 :      inst_sort=_STB_; /* Store Byte */
      byte = (unsigned char)(r[rd] & 0xff);
      memcpy((char *)mem_address(mem_adr), &byte, 1);
      break;
      
    case  6 :      inst_sort=_STH_; /* Store Halfword */
      halfwd = (unsigned short)(r[rd] & 0xffff);;
      memcpy((char *)mem_address(mem_adr), &halfwd, 2);
      break;
      
    case  7 :      inst_sort=_STD_; /* Store Doubleword */
      memcpy((char *)mem_address(mem_adr), &r[rd], 8);
      break;
      
    case 36 :      inst_sort=_STF_; /* Store Floating-Point register */
      memcpy((char *)mem_address(mem_adr), &f[rd], 4);   
      break;
      
    case 37 :    inst_sort=_STFSR_; /* Store Floating-Point Sta1te Register */
      memcpy((char *)mem_address(mem_adr), &fsr, 4);
      break;
      
    case 39 :      inst_sort=_STDF_; /* Store Doubleword Floating-Point */
      memcpy((char *)mem_address(mem_adr), &f[rd], 8);
      break;
      
    case 15 :      inst_sort=_SWAP_; /* Swap r register with memory */
      memcpy(&word_swap, (char *)mem_address(mem_adr), 4);
      temp_swap = r[rd];
      r[rd] = word_swap;
      memcpy((char *)mem_address(mem_adr), &temp_swap, 4);
      break;
      
    default:
      fprintf(stderr,"ld/st default in ld_st.cc\n"); exit(1);
    }
  return 0;
}

int transfer(){
  unsigned int op2;
  
  op2=(ir & OP2)>>22;
  switch(op2){
  case 0:{
    inst_sort=_UNIMP_;
    printf("TC:%d ,UNIMP is executed at pc:0x%x\n", total_c, pc);
    exit(0);
  }
  case 4:SETHI(ir); break;
  case 2:Bicc(ir);  break;
  case 6:FBfcc(ir); break;
  case 7:CBccc();   break;
  }
}

/************************************************************/
void main_loop(int argc, char **argv){
  unsigned int pc_tmp;
  total_c=0 ;
  toend=0;
  annul=0;
  
  while(!toend){
    ir = load_word(pc);
    r[0]=0;          /* initialize global 0  */
    
    inst_sort = 0;   /* initialize instruction */
    pc_tmp = pc;
    
    value1=value2=0;  /* for debug */
    
    if(end_tc!=0 && total_c>=end_tc){
      printf("\n%lld inst executed. and finish.\n",total_c);
      break;
    }
    
    if(annul == 0){
      if(option_flag & DEBUG_FLAG) debugmode(); 
      
      switch( (ir & OP)>>30 ){
      case 0: transfer(); break;
      case 1:	call();     break;
      case 2:	ar_log();   break;
      case 3:	ld_st();    break;
      }
      
      if(inst_sort!=_CALL_ && inst_sort!=_RETT_ &&
	 inst_sort!=_JMPL_ && inst_sort!=_Bicc_ &&
	 inst_sort!=_FBfcc_ && inst_sort!=_CBccc_){
	pc = npc;
	npc +=4;
      }
      
      total_c++;
    }
    else{
      annul = 0;
      pc = npc;
      npc +=4;
    }
  }
}

int main(int argc, char **argv, char **argp){
  construct_data(argc, argv, argp);
  main_loop(argc, argv);
  destruct_data(argc, argv);
  return 0;
}
